#include "GenAlg.h"
#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <random>
#include "OptScheduling.h"
#include "FirstOptScheduling.h"

GenAlg::GenAlg(std::shared_ptr<OptScheduling> optScheduling)
    :mOptScheduling(optScheduling)
{
  
}

double GenAlg::random(int min , int max)
{
    double m1 = (double)(rand() % 101) / 101;                        //  01֮С,õֵΪ(0,1)
    min++;                                                                             // Ϊ(min+1,max),
    double m2 = (double)((rand() % (max - min + 1)) + min);    // min+1,max ֮õֵΪ[min+1,max]
    m2 = m2 - 1;                                                                        //ֵΪ[min,max-1]
    return m1 + m2;                                                                //ֵΪ(min,max),Ϊ
}

void GenAlg::init(int popsize, double MutRate, double CrossRate, int GenLength, double LeftPoint, double RightPoint)
{
    mPopSize = popsize;
    mMutationRate = MutRate;
    mCrossoverRate = CrossRate;
    mChromoLength = GenLength;
    mTotalFitness = 0;
    mGeneration = 0;

    mBestFitness = 0.0;
    mWorstFitness = 99999999;
    mAverageFitness = 0;
    mMaxPerturbation = 0.001;

    mLeftPoint = LeftPoint;
    mRightPoint = RightPoint;

    // ȺԳʼ  
    vecPop.clear();
    for (int i = 0; i < mPopSize; i++)
    {
        // Ĺ캯ѾӦֳʼΪ0  
        vecPop.push_back(Genome());

        // еĻʼΪڵ  
        for (int j = 0; j < mChromoLength; j++)
        {
                vecPop[i].vecGenome.push_back(random(0,MAX_RANDOM) * (mRightPoint - mLeftPoint) + mLeftPoint);   
        }
    }
}

void GenAlg::mutate(vector<double>& chromo)
{
    // ѭԤͻ,Իͻ  
    for (int i = 0; i < chromo.size(); ++i)
    {
       // ͻĻ  
        if (random(0,1) < mMutationRate)
        {
            // ʹȨֵӻ߼һСֵ  
            chromo[i] += ((random() - 0.05) * mMaxPerturbation);

            // ұ߽  
            if (chromo[i] < mLeftPoint)
            {
                chromo[i] = mRightPoint;
            }
            else if (chromo[i] > mRightPoint)
            {
                chromo[i] = mLeftPoint;
            }
        }
    }
}

void GenAlg::select(vector<Genome>& genVec)
{
    // һ 0 ȺӦ֮ܺ.  mTotalFitness¼ȺӦԷܺ
    double slice = (random(0, 1)) * mTotalFitness;

    // 򽫳תѡǸ.  
    Genome theChosenOne;

    // ۼӦԷĺ.  
    double fitnessSoFar = 0;

    // ȺÿһȾɫ塣  
    for (int i = 0; i < mPopSize; ++i)
    {
        //ۼӦԷ.  
        fitnessSoFar += vecPop[i].fitness;

        // ۼƷ,ѡʱĻ.  
        if (fitnessSoFar >= slice)
        {
            theChosenOne = vecPop[i];
            genVec.push_back(theChosenOne);
            continue;
        }
    }
}

Genome GenAlg::chromoRoulette()
{
    // һ 0 ȺӦ֮ܺ.  mTotalFitness¼ȺӦԷܺ
    double slice = (random(0,1)) * mTotalFitness;

    // 򽫳תѡǸ.  
    Genome theChosenOne;

    // ۼӦԷĺ.  
    double fitnessSoFar = 0;

    // ȺÿһȾɫ塣  
    for (int i = 0; i < mPopSize; ++i)
    {
        //ۼӦԷ.  
        fitnessSoFar += vecPop[i].fitness;

        // ۼƷ,ѡʱĻ.  
        if (fitnessSoFar >= slice)
        {
            theChosenOne = vecPop[i];
            break;
        }
    }
    //תѡĸ  
    return theChosenOne;
}

void GenAlg::calculateBestWorstAvTot()
{
    int total = vecPop.size();
    for (int i = 0; i < total; i++)
    {
        double fitNessTemp = vecPop[i].fitness;

        // Ӧֵ
        mTotalFitness += fitNessTemp;

        // Ӧֵ
        if (mBestFitness < fitNessTemp)
            mBestFitness = fitNessTemp;

        // Ӧֵ
        if (mWorstFitness > fitNessTemp)
            mWorstFitness = fitNessTemp;
    }

    // ƽӦ
    mAverageFitness = mTotalFitness / (double)total;
}

void GenAlg::reset()
{
    mPopSize = 0;
    mMutationRate = 0;
    mCrossoverRate =0;
    mChromoLength = 0;
    mTotalFitness = 0;
    mGeneration = 0;

    mBestFitness = 0.0;
    mWorstFitness = 99999999;
    mAverageFitness = 0;
    mMaxPerturbation = 0.001;

    mLeftPoint =0;
    mRightPoint = 0;

    // ȺԳʼ  
    vecPop.clear();
}

void GenAlg::epoch(vector<Genome>& vecNewPop)
{
    // 㵱ǰȺӦֵ
    calculateBestWorstAvTot();

    // ݸӦȽѡ
    select(vecNewPop);
}

Genome GenAlg::bestFitness()
{
    int maxIndex = -1;
    for (int i = 0; i < vecPop.size(); i++)
    {
        if (mBestFitness == vecPop[i].fitness)  
            maxIndex = i;    
    }

    // ûҵһյĻͶ
    if (maxIndex == -1)
        return Genome();

    return vecPop[maxIndex];
}

double GenAlg::averageFitness()
{
    return mAverageFitness;
}

double  GenAlg::fitnessfunction(float cost, map<string, double>& monitorValues)
{
    // Ӧ
    double objVal = mOptScheduling->objectiveFunction(cost, monitorValues);
    return objVal;
}

void GenAlg::decoding(vector<map<string, float>>& genVals)
{
    size_t popSize = vecPop.size();
    for (int i = 0; i < popSize; i++)
    {
        Genome genome = vecPop[i];
        vector<double> genomeValsVec = genome.vecGenome;
        size_t genSize = genomeValsVec.size();

        map<string, float> genMap;
        for (int j = 0; j < genSize; i++)
        {
            genMap.insert(pair<string,float>(mPumpVec[j], genomeValsVec[j]));
        }
        genVals.push_back(genMap);
    }
}

void GenAlg::updateGenomeFitNess(int index, double fitNess)
{
    if (index < 0 || index >= vecPop.size())
        return;

    vecPop[index].fitness = fitNess;
}

void GenAlg::crossover(const vector<Genome>& parent, vector<Genome>& son)
{
    size_t pGenSize = parent.size();

    // Ⱥ
    std::vector<int> randIndexVec;
    for (int i = 0; i < pGenSize; i++)
        randIndexVec.push_back(i);

    // ΪȾɫ
    std::random_shuffle(randIndexVec.begin(), randIndexVec.end());

    int left = 0;
    int right = pGenSize - 1;
    while (left < right)
    {
        int indexLeft = randIndexVec[left];
        int indexRight = randIndexVec[right];
        
       Genome genome = crossover(parent[indexLeft], parent[indexRight]);
       son.push_back(genome);

        left++;
        right--;
    }
}

Genome GenAlg::crossover(const Genome& gen1, const Genome& gen2)
{
    Genome geoMe;

    vector<double> genVec1 = gen1.vecGenome;
    vector<double> genVec2 = gen2.vecGenome;

    if (genVec1.size() != genVec2.size())
        return geoMe;

    size_t genSize = genVec1.size();
    double a = 0.15;
    for (int i = 0; i < genSize; i++)
    {
        double newGenVal = genVec1[i] + a * (genVec1[i] - genVec2[i]);
        geoMe.vecGenome.push_back(newGenVal);
    }

    return geoMe;
}