#include "CivOptSchedEngine.h"
#include "CivInpConvertor.h"
#include "GenAlg.h"
#include "FirstOptScheduling.h"
#include "CivPumpHelper.h"
#include "Genome.h"
#include "CivCommonUtils.h"

CivOptSchedEngine::CivOptSchedEngine(const string& uri)
	:mUri(uri),mMutRate(0.45),mCrossRate(0.07), popsize(10)
{
	
}

bool CivOptSchedEngine::optimalScheduling(int time)
{
	// ʼ
	beginCompute();

	// ʼŴ㷨
	std::shared_ptr<OptScheduling> optScheduling(new FirstOptScheduling());
	optScheduling->setMinMonitorsVals(mMinMonitorVals);
	optScheduling->setMaxMonitorVals(mMaxMonitorVals);

	GenAlg genAlg(optScheduling);

	// Ⱥʼ
	popsize = 10;
	genAlg.init(popsize, mMutRate, mCrossRate, mGenLength, 0, 1);
	
	// ʼ
	int index = 0;

	// 
	int generation = 20;

	while (index++ < generation)
	{
		// ƽ
		vector<map<string, float>>  monitorsMap;
		genAlg.decoding(monitorsMap);

		// Ⱥÿ
		size_t mSize = monitorsMap.size();
		for (int i = 0; i < mSize; i++)
		{
			mSchedulingCompute->updatePumpSpeed(monitorsMap[i]);
			if (!mSchedulingCompute->calculate())
				continue;

			// ȡļֵ
			map<string, double> monitorCalcMapValues;
			mSchedulingCompute->getMonitorsValue(mPeriod, monitorCalcMapValues);

			// ȡĵֵ
			float cost = mSchedulingCompute->totalEnergy(mPeriod);
			
			// ʶ
			double fitness = genAlg.fitnessfunction(cost, monitorCalcMapValues);

			// ¸Ӧ
			genAlg.updateGenomeFitNess(i, fitness);
		}

		
		// 
		// genAlg.mutate()
		// һȺ
		std::vector<Genome> newGenmeVec;
		genAlg.epoch(newGenmeVec);
	}

	// ȡõ
	Genome genome = genAlg.bestFitness();
	vector<double> genVec = genome.vecGenome;

	map<string, float> genMap;
	for (int i = 0; i < mMonitorsVec.size(); i++)
		genMap.insert(pair<string, double>(mMonitorsVec[i], genVec[i]));

	// һˮ㣬Ⱥֵ
	mSchedulingCompute->updatePumpSpeed(genMap);
	mSchedulingCompute->calculate();

	// еֵ
	mSchedulingCompute->getMonitorsValue(mPeriod,mAfterSchedulingResults);

	return true;
}

void CivOptSchedEngine::schedulingResultToJson(string& json)
{
	// ˮ:
	json.append("{\"ˮ״̬\":{");

	map<string, int>::iterator iter = mPumpBestStatus.begin();
	for (;iter!=mPumpBestStatus.end();iter++)
	{
		json.append(iter->first + ":" + to_string(iter->second));
		json.append(",");
	}
	json = json.substr(0, json.length() - 1);
	json.append("},");

	// 
	json.append("\":[\"");
	int total = mBforeSchedulingResults.size();
	for (int i = 0; i < total; i++)
	{
		map<string, double> beforeResult = mBforeSchedulingResults[i];
		map<string, double> afterResult = mAfterSchedulingResults[i];

		json.append("{");
		for (auto iter = beforeResult.begin(); iter != beforeResult.end(); iter++)
		{
			string monitorName = iter->first;
			double beforeMonitorVal = iter->second;

			auto afteriter = afterResult.find(monitorName);
			if (afteriter == afterResult.end())
				continue;

			double afterMonitorvVal = afteriter->second;

			json.append("\""+monitorName+"\":[");
			json.append(to_string(beforeMonitorVal));
			json.append(",");
			json.append(to_string(afterMonitorvVal));
			json.append("],");
		}
		json = json.substr(0, json.length() - 1);
		json.append("},");
	}
	json.substr(0, json.length() - 1);
	json.append("]}");
}

void CivOptSchedEngine::updateMonitorsVals(const map<string, double>& monitorVals)
{
	mCurrentMonitorVals = monitorVals;
}

bool CivOptSchedEngine::beginCompute()
{
	// תinpļ
	CivInpConvertor convertor(mUri);
	string inpFile = convertor.convertBaseInp();

	// ʼ
	mSchedulingCompute = std::make_shared<CivSchedulingCompute>(CivSchedulingCompute::PumpScheduling);
	
	// ˮ
	mSchedulingCompute->setMonitors(mMonitorsVec);
	mSchedulingCompute->openFile(inpFile);

	// ԭʼһˮǰˮ
	if (!mSchedulingCompute->calculate())
		return false;

	// ԭʼģֵ
	mSchedulingCompute->getMonitorsValue(mPeriod,mBforeSchedulingResults);

	// ȡˮø,ˮøȷ볤
	CivPumpHelper pumpHelper(mUri);
	mGenLength = pumpHelper.getPumpNumbers();

	// ȡźcodeӳ
	map<string, string> snToCodeMap;
	pumpHelper.getMap("", "code", snToCodeMap);

	// ȡż
	vector<string> snVec;
	pumpHelper.getPumpSn(snVec);
}

void CivOptSchedEngine::endCompute()
{
	mmSchedulingCompute->close();
}

void CivOptSchedEngine::setMonitors(const vector<string>& monitor)
{
	mMonitorsVec = monitor;
}
